#!/usr/bin/env python2
import sys, time, argparse, math, json, os, random

autoselectnewpoints = True
newpolysavgcolor = True

sys.stderr = open('log.txt', 'a')
print >> sys.stderr, '***', time.ctime()


formatversion = 2

CLICKDIST = 10
DRAGDIST = 5

if len(sys.argv) > 1:
    ap = argparse.ArgumentParser()
    ap.add_argument('image')
    args = ap.parse_args()
    imagepath = os.path.abspath(args.image)
else:
    import Tkinter, tkFileDialog
    root = Tkinter.Tk()
    root.withdraw()
    imagepath = tkFileDialog.askopenfilename()
if not imagepath or not os.path.exists(imagepath):
    print >> sys.stderr, 'invalid file:', imagepath
    exit(1)

import pygame
pygame.init()
screen = pygame.display.set_mode((1024, 768))
image = pygame.image.load(imagepath).convert_alpha()
pygame.display.set_caption('lowpoly editor')
font = pygame.font.Font(pygame.font.get_default_font(), 18)

camera = (image.get_width() / 2., image.get_height() / 2.)
zoom = screen.get_width() / float(image.get_width())

sizedimage = None
sizedzoom = None
sizedpos = None
def drawbg():
    global sizedimage, sizedzoom, sizedpos
    if zoom != sizedzoom:
        if zoom <= 1:
            size = intpos((image.get_width() * zoom, image.get_height() * zoom))
            sizedimage = pygame.transform.smoothscale(image, size)
            sizedzoom = zoom
            sizedpos = (0, 0)
        else:
            a = screen2image((0, 0))
            b = screen2image(screen.get_size())
            r = pygame.Rect((a[0], a[1], b[0] - a[0], b[1] - a[1]))
            rc = image.get_rect().clip(r)
            size = intpos((rc.w * zoom, rc.h * zoom))
            sizedimage = pygame.transform.smoothscale(image.subsurface(rc), size)
            sizedzoom = (zoom, camera)
            sizedpos = (rc.x, rc.y)

    screen.blit(sizedimage, intpos(image2screen(sizedpos)))

def screen2image(pos):
    w, h = screen.get_size()
    return ((pos[0] - w/2) / zoom + camera[0], (pos[1] - h/2) / zoom + camera[1])

def image2screen(pos):
    w, h = screen.get_size()
    return ((pos[0] - camera[0]) * zoom + w/2, (pos[1] - camera[1]) * zoom + h/2)

def scale2image(n):
    return n / zoom

def intpos(pos):
    return (int(pos[0]), int(pos[1]))

class Point:
    def __init__(self, pos):
        self.x = pos[0]
        self.y = pos[1]
        
    @property
    def pos(self):
        return intpos(image2screen((self.x, self.y)))

class Poly:
    def __init__(self):
        self.pts = []
        self.color = (255, 255, 255)

    @property
    def poslist(self):
        return [pt.pos for pt in self.pts]

    @property
    def poslistrawint(self):
        return [intpos((pt.x, pt.y)) for pt in self.pts]

    @property
    def poslistrawfloat(self):
        return [(pt.x, pt.y) for pt in self.pts]

def showmsg(s):
    'for showing a message while doing something computationally intense'
    global dirty
    s = font.render(s, True, (0, 0, 0))
    r = (0, 0, screen.get_width(), s.get_height() * 2)
    screen.fill((255, 255, 255), r)
    screen.blit(s, (s.get_height() / 2, s.get_height() / 2))
    pygame.display.update([r])
    dirty = True

def saveall():
    save(filename)
    saveimage(fileroot + '_poly.png')
    savevector(fileroot + '_poly.svg')

def save(filename):
    showmsg('Saving to %s...' % os.path.basename(filename))
    
    delbackup = False
    if os.path.exists(filename):
        os.rename(filename, filename + 'backup')
        delbackup = True
        
    ptindex = {pt: i for i, pt in enumerate(points)}
    f = open(filename, 'w')
    d = {
        'version': formatversion,
        'points': [(pt.x, pt.y) for pt in points],
        'polys': [{'color': p.color, 'pts': [ptindex[pt] for pt in p.pts]} for p in polys]
    }
    json.dump(d, f, sort_keys=True, separators=(',',':'))
    f.close()
    
    if delbackup:
        os.remove(filename + 'backup')

def load(filename):
    f = open(filename, 'r')
    d = json.load(f)
    f.close()

    fileversion = d.get('version', 1)
    if fileversion != formatversion:
        print >> sys.stderr, 'FATAL: %s has format version %d (expecting %d)' % (filename, fileversion, formatversion)
        exit(1)

    for pos in d['points']:
        points.append(Point(pos))
    for poly in d['polys']:
        p = Poly()
        p.color = tuple(poly['color'])
        p.pts = [points[i] for i in poly['pts']]
        polys.append(p)

def render():
    screen.fill((255, 0, 255))
    if not hidepic:
        drawbg()
    
    for poly in polys:
        pygame.draw.polygon(screen, poly.color, poly.poslist, 2 if wireframe else 0)

    if showpoints:
        for pt in points:
            if pt in newpoly.pts:
                if len(newpoly.pts) == 1:
                    color = (0, 0, 255)
                else:
                    color = (255, 0, 0)
            else:
                color = (0, 255, 0)
            pygame.draw.circle(screen, color, pt.pos, 3)
        
    if selpoly != None:
        pygame.draw.polygon(screen, (0, 0, 255), selpoly.poslist, 2)
        
        pygame.draw.line(screen, (255, 0, 0), (10, 10), (10, 110), 5)
        c = 10 + 100 - selpoly.color[0] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (10, c), (15, c), 2)
        
        pygame.draw.line(screen, (0, 255, 0), (30, 10), (30, 110), 5)
        c = 10 + 100 - selpoly.color[1] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (30, c), (35, c), 2)
        
        pygame.draw.line(screen, (0, 0, 255), (50, 10), (50, 110), 5)
        c = 10 + 100 - selpoly.color[2] * 100 / 255
        pygame.draw.line(screen, (0, 0, 0), (50, c), (55, c), 2)
        
    pygame.display.update()

def saveimage(filename):
    showmsg('Rendering to %s...' % os.path.basename(filename))
    
    surf = pygame.Surface(image.get_size(), pygame.SRCALPHA)
    surf.fill((0, 0, 0, 0))

    for poly in polys:
        pygame.draw.polygon(surf, poly.color, poly.poslistrawint)
        
    pygame.image.save(surf, filename)

def savevector(filename):
    showmsg('Saving vectors to %s...' % os.path.basename(filename))

    fh = open(filename, 'w')
    w, h = image.get_size()
    # magic svg header based on an example at http://www.w3.org/TR/SVG11/paths.html
    print >> fh, '''\
<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg width="%d" height="%d" viewBox="0 0 %d %d" xmlns="http://www.w3.org/2000/svg" version="1.1">
<style> polygon { stroke-width: .5; stroke-linejoin: round; } </style>
''' % (w, h, w, h)

    for poly in polys:
        plst = ' '.join('%.3f,%.3f' % (pt[0], pt[1]) for pt in poly.poslistrawfloat)
        color = 'rgb(%d,%d,%d)' % poly.color
        print >> fh, '<polygon style="fill:%s;stroke:%s" points="%s" />' % (color, color, plst)

    print >> fh, '</svg>'
    fh.close()

def getcolorat(pos):
    try:
        return image.get_at(intpos(pos))[:3]
    except IndexError:
        return (255, 0, 255)

def getnearest(pos, maxdist=None):
    best = None
    bestpt = None
    for pt in points:
        d = math.hypot(pt.x - pos[0], pt.y - pos[1])
        if (maxdist == None or d <= maxdist) and (best == None or d < best):
            best = d
            bestpt = pt
    return bestpt

def getpolycenter(poly):
    return (sum(pt.x for pt in poly.pts) / 3, sum(pt.y for pt in poly.pts) / 3)

def getnearestpoly(pos):
    best = None
    bestpoly = None
    for poly in polys:
        center = getpolycenter(poly)
        d = math.hypot(center[0] - pos[0], center[1] - pos[1])
        if best == None or d < best:
            best = d
            bestpoly = poly
    return bestpoly

def colorwidgetpos(pos):
    '''
    return (x, y) or None
    x will be in [0, 1, 2]
    y will be between 0 and 1
    '''
    x, y = image2screen(pos)
    if x < 60 and y < 120:
        x = int(x / 20)
        y = 1 - (y - 10) / 100.
        if y < 0:
            y = 0
        elif y > 1:
            y = 1
        return (x, y)
    return None

def startdragging(button, pos):
    global movepoint
    if button == 1:
        movepoint = getnearest(mousedown[0], CLICKDIST)

def dragging(button, pos, rel):
    global movepoint, dirty, camera

    if button == 1:
        # dragging in color widget
        if selpoly != None and colorwidgetpos(mousedown[0]):
            if colorwidgetpos(pos):
                click(pos)
            return

        if movepoint != None:
            movepoint.x += rel[0]
            movepoint.y += rel[1]
            dirty = True

    elif button == 2:
        camera = (camera[0] - rel[0], camera[1] - rel[1])
        dirty = True

def selectpoint(pt):
    global dirty, selpoly, newpoly
    if pt not in newpoly.pts:
        newpoly.pts.append(pt)
        if len(newpoly.pts) == 3:
            if newpolysavgcolor:
                randavgcolorpoly(newpoly)
            else:
                newpoly.color = getcolorat(getpolycenter(newpoly))
            selpoly = newpoly
            polys.append(newpoly)
            newpoly = Poly()
        dirty = True

def click(pos):
    global points, newpoly, dirty, selpoly, showpoints

    if selpoly != None and colorwidgetpos(pos):
        i, cf = colorwidgetpos(pos)
        c = list(selpoly.color)
        c[i] = 256 * cf
        if c[i] < 0:
            c[i] = 0
        elif c[i] > 255:
            c[i] = 255
        selpoly.color = tuple(c)
        dirty = True
        return
    
    if len(points) > 0:
        selpoly = None
        pt = getnearest(pos, scale2image(CLICKDIST))
        if pt != None and pt not in newpoly.pts:
            selectpoint(pt)
            return

    pt = Point(pos)
    points.append(pt)
    if autoselectnewpoints:
        selectpoint(pt)
    showpoints = True
    dirty = True

def scrollwheel(dir):
    global zoom, dirty
    if dir > 0:
        zoom *= 1.1
    else:
        zoom /= 1.1
    dirty = True

def rightclick(pos):
    global selpoly, dirty
    if len(polys) > 0:
        selpoly = getnearestpoly(pos)
        dirty = True

def randpointinpoly(poly):
    x = random.random()
    y = random.random()
    if x + y >= 1:
        x = 1 - x
        y = 1 - y

    pts = poly.poslistrawint
    a = pts[0]
    b = (pts[1][0] - a[0], pts[1][1] - a[1])
    c = (pts[2][0] - a[0], pts[2][1] - a[1])
    return intpos((a[0] + b[0]*x + c[0]*y, a[1] + b[1]*x + c[1]*y))

def randavgcolorpoly(poly, count=10):
    t = [0, 0, 0]
    for i in range(count):
        c = getcolorat(randpointinpoly(poly))
        for k in range(3):
            t[k] += c[k]
    poly.color = tuple(x / count for x in t)

def keypress(key):
    global newpoly, dirty, selpoly, wireframe, hidepic, showpoints
    
    if key == pygame.K_c:
        if selpoly != None:
            selpoly.color = getcolorat(getpolycenter(selpoly))
            dirty = True
    elif key == pygame.K_o:
        if selpoly != None:
            pos = screen2image(pygame.mouse.get_pos())
            selpoly.color = getcolorat(pos)
            dirty = True
    elif key == pygame.K_a:
        if selpoly != None:
            randavgcolorpoly(selpoly)
            dirty = True
            
    elif key == pygame.K_DELETE:
        if selpoly != None:
            polys.remove(selpoly)
            selpoly = None
            dirty = True
        elif len(newpoly.pts) > 0:
            for pt in newpoly.pts:
                points.remove(pt)
                for poly in polys[:]:
                    if pt in poly.pts:
                        polys.remove(poly)
            newpoly = Poly()
            dirty = True
            
    elif key == pygame.K_w:
        wireframe = not wireframe
        dirty = True
    elif key == pygame.K_h:
        hidepic = not hidepic
        dirty = True
    elif key == pygame.K_p:
        showpoints = not showpoints
        dirty = True
        
    elif key == pygame.K_s:
        saveall()
        
    elif key == pygame.K_MINUS:
        if selpoly != None:
            polys.remove(selpoly)
            polys.insert(0, selpoly)
            dirty = True
    elif key == pygame.K_EQUALS:
        if selpoly != None:
            polys.remove(selpoly)
            polys.append(selpoly)
            dirty = True
            
    elif key == pygame.K_SPACE:
        selpoly = None
        newpoly = Poly()
        dirty = True

dirty = True
running = True

points = []
polys = []
newpoly = Poly()
selpoly = None

wireframe = False
hidepic = False
showpoints = True

mousedown = None
mousedrag = False
movepoint = None

fileroot = os.path.splitext(imagepath)[0]
filename = fileroot + '.poly'
if os.path.exists(filename):
    load(filename)

while running:
    for event in pygame.event.get():
        if event.type == pygame.MOUSEBUTTONDOWN:
            pos = screen2image(event.pos)
            if event.button in (1, 2):
                mousedown = (pos, event.button)
            elif event.button == 3:
                rightclick(pos)
            elif event.button in (4, 5):
                scrollwheel(-1 if event.button == 5 else 1)
        elif event.type == pygame.MOUSEMOTION:
            pos = screen2image(event.pos)
            rel = (scale2image(event.rel[0]), scale2image(event.rel[1]))
            if mousedown and not mousedrag and math.hypot(pos[0] - mousedown[0][0], pos[1] - mousedown[0][1]) > scale2image(DRAGDIST):
                mousedrag = True
                startdragging(mousedown[1], mousedown[0])
            if mousedrag:
                dragging(mousedown[1], pos, rel)
        elif event.type == pygame.MOUSEBUTTONUP:
            if event.button == 1:
                if not mousedrag:
                    click(mousedown[0])
            mousedown = None
            mousedrag = False
            movepoint = None
        elif event.type == pygame.KEYDOWN:
            keypress(event.key)
        elif event.type == pygame.QUIT:
            running = False
    if dirty:
        render()
        dirty = False
    time.sleep(.01)

saveall()
